#!/usr/bin/env python
'''Creates an RPM spec file from a project makefile or file list

Copyright 1998-2001 Daniel Fandrich <dan@coneharvesters.com>

$Id: autospec.py,v 1.10 2001/05/08 20:45:13 dan Exp $
'''

# Bugs:
#  - if the name ends in digits (e.g. Alec64) the digits are interpreted as 
#    the version number
#
# To do:
#  - Do a regsub.sub on the source file name, substituting %{version} for the
#    version number
#  - add more documentation files to classifyfile.py (remember Slackware)
#  - add directory prefix option for -b for relative binary tgzs (e.g. maxwell)
#  - add './configure' line in %install if file exists in current directory
#  - improve command line parsing, especially quotes and redirection
#  - support for chmod -R
#  - if COPYING exists and it looks like GPL, set Copyright: line to GPL
#  - clean up parseinstall by splitting command parsing into separate functions
#    and create a class to hold them all
#  - maybe catch all redirections, not just those for echo
#  - in the case of file globbing in cp command, actually look at files on
#    disk to see what's there
#  - maybe handle if [ xxx ] ; then ... enough for xmkmf makefiles
#  - add option that takes files on stdin and creates an install script
#    that guesses where the non-absolute files go, either /usr/bin,
#    /usr/X11R6/bin, /usr/lib/{rpmname}, /usr/man/manX/
#    - guess based on file modes, new patterns, `ldd' dependencies, `file' type
#  - look at all files and enable 'BuildArchitectures: noarch' if applicable
#  - look for documentation file named 'foo.txt' where foo is the RPM name
#  - look at some of the other % file flags, like %config(missingok)
#  - support bzip2 source archives
# 
# $Log: autospec.py,v $
# Revision 1.10  2001/05/08 20:45:13  dan
# Added -z option for ZIP archives.  Added %clean section.  Added quoting
# of file names with special characters.  Fixed bug in BuildRoot name.
# Eliminated other file flags if %dir is present.
#
# Revision 1.9  2001/03/30 04:51:45  dan
# Added -r RPM version compatibility option.  Improved install script parsing
# and fixed exception that occurred with some input.  Added more magic
# documentation file names.  Using BuildRoot whenever -i not used. Added
# %changelog section. Created options class.  Upped version number to 0.5
#
# Revision 1.8  2000/06/05 19:51:57  dan
# Improved comand-line parsing by at least partially supporting quoted args.
#
# Revision 1.7  1999/03/27 20:09:17  dan
# Change to use new splitfirst semantics. Improve chown processing.
# Add findurl to look for URLs in the description field of the LSM file.
# Add install commands to the %install section if -i is not given.
# Upped version number to 0.3.1
#
# Revision 1.6  1999/02/26 08:12:47  dan
# Added ranlib to install ignore list. Upped version number to 0.3
#
# Revision 1.5  1999/02/26 07:15:10  dan
# Now using posixpath to handle file names. Renamed -m to -i.
# Added -m option to take information from LSM file.  Reversed meaning of -d.
# Added entries to doc name list. Added bzip2 to install ignore list.
#
# Revision 1.4  1998/08/06 04:20:07  dan
# Added -c and -g options. Upped version number to 0.2.
# 
# Revision 1.3  1998/07/20 20:47:42  dan
# Added -b and -n options.
# 
# Revision 1.2  1998/07/14 23:00:20  dan
# Added my name and e-mail address.
# 
# Revision 1.1  1998/07/14 22:52:16  dan
# Initial revision
#

license = '''\
   Copyright 1997-2001 by Daniel Fandrich <dan@coneharvesters.com>
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
'''

usage_msg = '''\
Automatically creates an RPM spec file with files given on stdin

usage: %s [-bdhimL] [-c copyright] [-g groupname] [-n name-version] 
          [-r rpm-version]
 -b         package source contains binaries in the proper directories
 -c         specifies copyright restrictions (Copyright:)
 -d         don't search for extra documentation files in current directory
 -g         specifies the classification group for the package (Group:)
 -h         show this help
 -i         install script (e.g. 'make -n install') appears on stdin 
 -m         don't use a Linux Software Map (LSM) file if found in current dir
 -n         RPM name and optional version (Name: and Version:)
 -r         produce spec file compatible with this RPM version
 -z         source is in a ZIP archive instead of a gzipped tar ball
 -L         display software license
'''

version_msg = 'autospec ver. 0.6'

from string import strip, split, join, find, atoi, atof, index
from os import error, path, getcwd, listdir, getuid, uname
from urlparse import urlparse
from pwd import getpwuid
import commands
import time
import getopt
import regex
import sys
import posixpath

from classifyfile import *
from danutil import printerr,splitfirst,quotedsplit
from lsm import *

OSError = error		# "error" is too generic, so rename and get rid of it
del error

defaultmode = 644		# default file mode
defaultexemode = 755	# default file mode for executables
defaultowner = 'root'	# default file owner
defaultownernum = '0'	# default file owner (numerical)
defaultgroup = 'root'	# default file group
defaultgroupnum = '0'	# default file group (numerical)

#debuglog = printerr			# enable this for debug logs
def debuglog(errstr): pass		# enable this to stop debug logs

class Options:
	'Class containing all the spec file generation options'
	docfilepath = '.'
	rpmnamever = None
	group = None
	copyright = None
	binarymode = 0
	zipmode = 0
	uselsm = 1
	rpmver = 2.5

class FileList:
	'List-like class to hold information about files'
	class FileInfo:
		'Structure for holding information about an installed file'
		name = None
		isdir = None
		mode = None
		owner = None
		group = None

	filelist = []

	def __len__(self):
		return len(self.filelist)

	def __getitem__(self,index):
		return self.filelist[index]

	def isadir(self, name):
		'Returns 1 if a file is in the list and is marked as a directory'
		file = self.find(name)
		if file:
			return file.isdir
		return 0

	def has_key(self, name):
		if self.find(name):
			return 1

	def find(self, name):
		'Returns FileInfo entry if a file is in the list'
		for file in self.filelist:
			if file.name == name:
				return file
		return None

	def append(self, info):
		'Add a FileInfo class to the file list'
		self.filelist.append(info)

	def getfilelist(self):
		'Return entire file list structure'
		return self.filelist

	def filenames(self):
		'Return list of all file names'
		return map(lambda f: f.name, self.filelist)

def betternormpath(name):
	'Normalize path, removing duplicated leading /'
	normname = posixpath.normpath(name)
	while normname[0:2] == '//':
		normname = normname[1:]
	return normname

def betterisabs(name):
	'See if a path is absolute, including relative to a specific user'
	return posixpath.isabs(name) or (name and name[0] == '~')

def abspath(name):
	'Make path absolute'
	if not posixpath.isabs(name):
		name = posixpath.expanduser(name)
	if not posixpath.isabs(name):
		name = betternormpath(posixpath.join(getcwd(),name))
	return betternormpath(name)


def multiplefiles(args, filelist, mode=None, owner=None, group=None):
	'''Handle potentially many files being copied (or whatever) to a single
	target'''

	sourcewild = 0
	if find(args[0], '*') >= 0 or find(args[0], '?') >= 0:
		sourcewild = 1		# destination must be a directory
	target = args[-1]

	if len(args) > 2 or sourcewild or filelist.isadir(target) \
	   or lookslikedir(target) or posixpath.isdir(target):
		# target is a directory
		# mark it as such in the file list
		debuglog('TARGET IS DIR: %s' % target)
		targetentry = filelist.find(target)
		if targetentry:
			targetentry.isdir = 1

		for source in args[:-1]:
			source = posixpath.basename(source)
			file = betternormpath(posixpath.join(target, source))
			fileinfo = filelist.FileInfo()
			fileinfo.name = abspath(file)
			fileinfo.mode = mode
			fileinfo.owner = owner
			fileinfo.group = group
			filelist.append(fileinfo)
	else:
		# target is a file
		debuglog('TARGET IS FILE: %s' % target)
		fileinfo = filelist.FileInfo()
		fileinfo.name = betternormpath(target)
		fileinfo.mode = mode
		fileinfo.owner = owner
		fileinfo.group = group
		filelist.append(fileinfo)


def parseinstall(script,filelist):
	'''Parses the commands in script to find what files are copied

	Inputs: file containing install script
	        filelist class
	Returns: number of errors while parsing script
	'''
	argserr = 'Warning: invalid number of arguments to %s'
	argsopterr = 'Warning: invalid number of arguments or unknown option to %s'
	ignorecmds = [
		# compile commands unrelated to installing
		'cc','gcc','g++','yacc','byacc','bison','lex','flex',
		# other commands to ignore
		'cd','ranlib','strip','rm','gzip','compress','bzip2'
	]

	errs = 0
	lineremaining = ''
	while 1:
		if lineremaining:
			# Process the remainder of a line after a semicolon
			line = lineremaining
			lineremaining = ''
		else:
			line = script.readline()
			if not line:
				break

		# Concatenate lines ending in \
		while line[-2:] == '\\\n':
			newline = script.readline()
			if not newline:
				break
			line = line[:-2] + newline

		# Break line up at the ; command separator
		# This is very broken as it ignores quotes and escape sequences
		try:
			line,lineremaining = splitfirst(line,';')
			debuglog('Remaining line: '+`lineremaining`)
		except ValueError:
			pass

		# This is doing very simple-minded parsing for now
		# This is also very broken for the same reasons
		tokens = quotedsplit(strip(line))
		debuglog('Input tokens: ' + `tokens`)

		if not tokens[0]:
			# It's a blank line--ignore it
			continue

		if tokens[0][0] == '#':
			# It's a comment--ignore it
			continue

		if tokens[0][0] == '(':
			errs = errs + 1
			printerr('Warning: subshells not supported')
			continue

		if find(line, '|') >= 0:
			errs = errs + 1
			printerr('Warning: command pipes not supported')
			continue

		# Strip any leading path names on the command
		command = posixpath.basename(tokens[0])
		args = tokens[1:]
		debuglog("Handling '%s'" % command)

		hasoptions = 0
		if args and args[0] and args[0][0] == '-':
			hasoptions = 1	# command line option was found

		# Ignore any commands in the ignore list
		try:
			ignorecmds.index(command)
			ignore = 1
		except ValueError:
			ignore = 0
		if ignore:
			continue

		# Process each install command, extracting the name of each
		# file installed.
		if command == 'cp':
			if len(args) < 2:
				errs = errs + 1
				printerr(argserr % command)
				continue

			if args[0] == '-f':
				# ignore -f option
				args = args[1:]

			if len(args) < 2 or args[0][0] == '-':
				errs = errs + 1
				printerr(argsopterr % command)
				continue

			target = args[-1]
			if not posixpath.isabs(target):
				# Ignore copies of files into the current directory
				continue

			multiplefiles(args, filelist)

		elif command == 'install' or command == 'ginstall':
			if len(args) < 2:
				errs = errs + 1
				printerr(argserr % command)
				continue

			try:
				optlist, args = getopt.getopt(args, 'cdsg:m:o:')
			except getopt.error:
				printerr('Unsupported option given to install')
				continue

			# Look at all options
			installdir = 0
			mode = owner = group = None
			for opt,arg in optlist:
				if opt == '-c':
					# ignore
					pass
				elif opt == '-s':
					# ignore
					pass
				elif opt == '-m':
					# setting file mode
					mode = atoi(arg)
				elif opt == '-o':
					# setting file owner
					owner = arg
				elif opt == '-g':
					# setting file group
					group = arg
				elif opt == '-d':
					# create a directory
					installdir = 1

			if installdir:
				# Installing (creating) a directory
				if len(args) != 1:
					errs = errs + 1
					printerr(argserr % command)
					continue
				fileinfo = filelist.FileInfo()
				fileinfo.name = abspath(args[0])
				fileinfo.isdir = 1
				fileinfo.mode = mode
				filelist.append(fileinfo)
				continue

			# Installing files
			if len(args) < 2:
				errs = errs + 1
				printerr(argserr % command)
				continue

			target = args[-1]

			if not posixpath.isabs(target):
				# Ignore copies of files into the current directory
				continue

			multiplefiles(args, filelist, mode, owner, group)

		elif command == 'touch':
			if len(args) != 1 or hasoptions:
				errs = errs + 1
				printerr(argsopterr % command)
				continue

			target = args[0]
			if not betterisabs(target):
				# Ignore touch of file in the current directory
				continue
			fileinfo = filelist.FileInfo()
			fileinfo.name = abspath(target)
			if lookslikedir(target) or posixpath.isdir(target):
				fileinfo.isdir = 1
			filelist.append(fileinfo)

		elif command == 'chown':
			if len(args) != 2 or hasoptions:
				errs = errs + 1
				printerr(argsopterr % command)
				continue

			target = args[-1]
			if not betterisabs(target):
				# Ignore chown of file in the current directory
				continue
			target = abspath(target)

			fileinfo = filelist.find(target)
			if not fileinfo:
				# new file
				fileinfo = filelist.FileInfo()
				fileinfo.name = target
				if lookslikedir(target) or posixpath.isdir(target):
					fileinfo.isdir = 1
				filelist.append(fileinfo)
			# Split off group if specified
			owners = split(args[0],'.')
			if len(owners) == 1:
				# Maybe group is preceded by a colon instead of a period
				owners = split(args[0],':')
			fileinfo.owner = owners[0]
			if len(owners) > 1:
				fileinfo.group = owners[1]

		elif command == 'chgrp':
			if len(args) != 2 or hasoptions:
				errs = errs + 1
				printerr(argsopterr % command)
				continue

			target = args[-1]
			if not betterisabs(target):
				# Ignore chgrp of file in the current directory
				continue
			target = abspath(target)

			fileinfo = filelist.find(target)
			if not fileinfo:
				# new file
				fileinfo = filelist.FileInfo()
				fileinfo.name = target
				if lookslikedir(target) or posixpath.isdir(target):
					fileinfo.isdir = 1
				filelist.append(fileinfo)
			fileinfo.group = args[0]

		elif command == 'chmod':
			if len(args) != 2 or hasoptions:
				errs = errs + 1
				printerr(argsopterr % command)
				continue

			target = args[-1]
			if not betterisabs(target):
				# Ignore chmod of file in the current directory
				continue
			target = abspath(target)

			fileinfo = filelist.find(target)
			if not fileinfo:
				# new file
				fileinfo = filelist.FileInfo()
				fileinfo.name = target
				if lookslikedir(target) or posixpath.isdir(target):
					fileinfo.isdir = 1
				filelist.append(fileinfo)

			try:
				fileinfo.mode = atoi(args[0])
			except ValueError:
				errs = errs + 1
				printerr('Warning: Invalid mode specified in chmod')
				# continue without the changed mode

		elif command == 'ln':
			try:
				optlist, args = getopt.getopt(args, 'cdsg:m:o:')
			except getopt.error:
				printerr('Unsupported option given to install')
				continue

			# Look at all options
			for opt,arg in optlist:
				if opt == '-b':
					# ignore
					pass
				elif opt == '-f':
					# ignore
					pass
				elif opt == '-s':
					# ignore
					pass
				elif opt == '-v':
					# ignore
					pass

			if len(args) != 2:
				errs = errs + 1
				printerr(argserr % command)
				continue

			target = args[-1]
			if not betterisabs(target):
				# Ignore link of file in the current directory
				continue

			multiplefiles(args, filelist)

		elif command == 'mkdir':
			if len(args) < 1:
				errs = errs + 1
				printerr(argserr % command)
				continue

			if args[0] == '-p':
				# ignore -p option
				args = args[1:]

			if len(args) != 1 or args[0][0] == '-':
				errs = errs + 1
				printerr(argsopterr % command)
				continue

			fileinfo = filelist.FileInfo()
			fileinfo.name = abspath(args[0])
			fileinfo.isdir = 1
			filelist.append(fileinfo)

		elif command == 'echo':
			# This test for redirection is far from perfect
			if find(line, '>') >= 0:
				errs = errs + 1
				printerr('Warning: redirected echo not supported')
				continue
			# ignore echo otherwise

		else:
			errs = errs + 1
			printerr("Warning: Unsupported command '%s' in install script" % command)
		# Some commands are deliberately not processed:
		#  mv - not usually used to install, but sometimes used to make
		#       backups of configuration files
		#  sh - sometimes used to run a script to do more configuration or
		#       or installation

	return errs


def readfilelist(filesfile,filelist):
	'''Read the list of files and put into the filelist class
	This function doesn't look at the current filesystem to get any info.
	'''
	errs = 0
	for fn in filesfile.readlines():
		f = filelist.FileInfo()
		f.name = strip(fn)
		if len(split(f.name)) > 1:
			errs = errs + 1
			printerr("Warning: doesn't look like a file name; try the -i option")
		if not posixpath.isabs(f.name):
			f.name = posixpath.join('/',f.name)
		if lookslikedir(f.name):
			f.isdir = 1
		filelist.append(f)
	return errs


# Patterns matching extra documentation files
# Don't forget to put complex regular expressions in parenthesis to prevent
# the special characters from applying to too much
# These are all case insensitive
doc_file_re = regex.compile(
              '^[0-]*\('			# This can apply to many of the following
                '\(READ[\._-]*ME.*\)\|'
				'ANNOUNCE\|'
				'AUTHOR\|'
				'BUGS\|'
				'CHANGES\|'
				'COPYING\|'
				'CREDITS\|'
				'ChangeLog\|'
				'Changes$\|'
				'Copyright\|'
				'\(docs?$\)\|'		# directory
				'documentation\|'	# directory
				'FAQ\|'
				'HISTORY$\|'
				'INSTALL$\|'
				'KNOWNBUGS\|'
				'License\|'
				'NEWS$\|'
				'NOTES$\|'
				'Problems\|'
				'QuickStart\|'
				'RELEASE\|'
				'RELNOTES\|'
				'TODO\|'
				'WHATSNEW'
				'\)\|\('			# Following are suffixes
				'\.html?$\|'
				'\.doc$\|'
				'\.txt$\|'
				'-LICENSE$'
				'\)'
				, regex.casefold)

def betterlookslikedocfile(file):
	'Returns 1 if a file has a common documentation file name'
	base = posixpath.basename(file)
	if lookslikedocfile(file) or doc_file_re.search(base) >= 0:
		return 1
	return 0

def finddocfiles(dir):
	'Search for documentation files in the given directory'
	docfiles = []
	for fn in listdir(dir):
		if betterlookslikedocfile(fn):
			docfiles.append(fn)
	docfiles.sort()
	return docfiles

def findlsmfiles(dir):
	'Search for .lsm files in the given directory'
	lsmfiles = filter(lambda fn: fn[-4:]=='.lsm', listdir(dir))
	lsmfiles.sort()
	return lsmfiles

# Look at the RFC and put all the legal characters here
url_re = regex.compile(
			  '\(\(http\|gopher\|ftp\|file\):'
			  '[a-z0-9~#%;/?:@=&$_.+!*''(),-]*\)'
			  , regex.casefold)

def findurl(str):
	'Search for what looks like a URL in a list of strings'
	for l in str:
		if url_re.search(l) >= 0:
			return url_re.group(1)
	return ''

def quotefilename(name):
	'Quote a file name if necessary to put it into the %files section'
	# Double quotes in a file name don't need additional quoting
	if regex.match('[a-zA-Z0-9\.+_=:/"-]*',name) != len(name):
		return '"'+name+'"'
	return name

def makearg(name):
	'Quote a file name if necessary to pass it to the shell'
	if regex.match('[a-zA-Z0-9\.+_=:/-]*',name) != len(name):
		return commands.mkarg(name)[1:]	# remove leading space
	return name

def install2spec(installscript,filesfile,outfile,options):
	'''Creates RPM spec file

	installscript - file containing installation commands
	filesfile - file containing list of files 
	outfile - write spec file here
	options - class containing configuration options
	'''
	unknown = 'unknown'	# string to use for unknown fields
	o = outfile

	# Use BuildRoot when we can
	buildroot = options.binarymode or not installscript

	lsm = {}
	if options.uselsm:
		lsmfiles = findlsmfiles('.')
		if lsmfiles:
			try:
				lsm = CanonicalLSM(open(lsmfiles[0],'r'))	# Use first name
			except:
				printerr("Can't use LSM file " + lsmfiles[0])
				lsm = {}

	if options.rpmnamever:
		dirname = options.rpmnamever	# user specified name
	else:
		dirname = path.basename(getcwd())
	version = extractver(dirname)
	if not version:
		if lsm.has_key('version'):
			version = joinstrings(lsm['version'])
		else:
			version = unknown

	# LSM Title: is often a long form name and shouldn't be used as the RPM name
	rpmname = dirname
	# if rpmname ends with the version string, strip it off
	if rpmname[-len(version):] == version and len(rpmname) > len(version):
		rpmname = rpmname[:-len(version)]
		dirname = rpmname + '%{version}'	# make spec file easier to update
	if rpmname[-1:] in ['-','_']:
		rpmname = rpmname[:-1]	# strip trailing - or _

	release = '1'
	distribution = 'Red Hat Contrib-Net'
	if not options.copyright:
		if lsm.has_key('copying-policy'):
			options.copyright = joinstrings(lsm['copying-policy'])
		else:
			options.copyright = unknown
	if not options.group:
		options.group = unknown

	parseerrs = 0
	filelist = FileList()
	if installscript:
		# Look at the installation commands and get the files from it
		parseerrs = parseinstall(installscript,filelist)
	if filesfile:
		# Read the list of files to use directly
		parseerrs = readfilelist(filesfile,filelist)

	files = filelist.filenames()
	prefix = dirprefix(files)

	# Find additional documentation files
	docfiles = []
	if options.docfilepath:
		docfiles = finddocfiles(options.docfilepath)

	# Start writing RPM spec file

	o.write('# Initial spec file created by %s with rpm %g compatibility\n'
			% (version_msg, options.rpmver))
	if lsm.has_key('title'):
		o.write('Summary: %s\n' % joinstrings(lsm['title']))
	else:
		o.write('Summary: %s\n' % rpmname)
		o.write('# The Summary: line should be expanded to about here -----^\n')

	o.write('Name: %s\n' % rpmname)		# required line

	if options.rpmver < 2.5:
		o.write('%%define version %s\n' % version)
		o.write('Version: %{version}\n')	# required line
	else:
		o.write('Version: %s\n' % version)	# required line

	o.write('Release: %s\n' % release)	# required line
	if options.group == unknown:
		o.write('#')
	o.write('Group: %s\n' % options.group)		# required line
	if options.copyright == unknown:
		o.write('#')
	o.write('Copyright: %s\n' % options.copyright)		# required line

	if options.binarymode:
		source = dirname + '-bin'
	else:
		source = dirname

	if options.zipmode:
		source = source + '.zip'
	else:
		source = source + '.tar.gz'

	if lsm.has_key('primary-site'):
		try:
			urls = fileurl(lsm['primary-site'])
			url = urls[0][0]
			if len(urls) > 1:
				# Choose a file that looks like an archive, if possible
				for u in urls:
					if find(u[0],'.tar') > 0 or \
					   find(u[0],'.tgz') > 0 or \
					   find(u[0],'.zip') > 0:
						url = u[0]
						break
			if find(posixpath.basename(urlparse(url)[2]),'.') < 0:
				# Looks like URL pointed to a directory, not an archive
				url = posixpath.join(url,source)
			source = url
		except BadSite:
			printerr("Badly-formatted Primary-Site: in LSM")
	o.write('Source: %s\n' % source)

	if not buildroot:
		o.write('#')	# BuildRoot in binary mode and without install script
	o.write('BuildRoot: /tmp/%s-root\n' % dirname)
	o.write('# Following are optional fields\n')

	url = ''
	if lsm.has_key('description'):
		url = findurl(lsm['description'])
	if url:
		o.write('URL: %s\n' % url)
	else:
		o.write('#URL: http://www.example.net/%s/\n' % rpmname)
	o.write('#Distribution: %s\n' % distribution)
	o.write('#Patch: %s.patch\n' % dirname)
	o.write('#Prefix: %s\n' % prefix)
	o.write('#BuildArchitectures: noarch\n')
	o.write('#Requires: \n')
	o.write('#Obsoletes: \n')
	#o.write('#BuildPreReq: \n')

	# End of header section
	o.write('\n')

	# Description is required
	# Should probably quote all % in description, but is this ever needed?
	o.write('%description\n')
	if lsm.has_key('description'):
		for l in lsm['description']:
			o.write('%s\n' % l)
	else:
		o.write('%s version %s\n' % (rpmname, version))
	o.write('\n')

	# Default prep commands
	o.write('%prep\n')
	if options.zipmode:
		o.write('%%setup -T -c %s\n' % dirname)
		o.write('unzip -aL ${RPM_SOURCE_DIR}/%s\n' % source)
	else:
		if options.binarymode:
			o.write('%%setup -c %s\n' % dirname)
		else:
			if version == unknown:
				o.write('%%setup -n %s\n' % dirname)
			else:
				o.write('%setup\n')
	o.write('#%patch\n')
	o.write('\n')

	# Default build and install commands
	if not options.binarymode:
		o.write('%build\n')
		o.write('#CFLAGS="$RPM_OPT_FLAGS" ./configure --prefix=/usr\n')
		o.write('make\n')
		o.write('\n')

	o.write('%install\n')
	if options.binarymode:
		o.write('cp -a . ${RPM_BUILD_ROOT-/}\n')
	else:
		if not installscript:
			createddirs = {'/':1}	# set of directories as we create each one
			for file in filelist.getfilelist():
				if not createddirs.has_key(path.dirname(file.name)):
					# Create the directory in the build root
					o.write('install -d ${RPM_BUILD_ROOT}%s\n' % 
							makearg(path.dirname(file.name)))
					createddirs[path.dirname(file.name)] = 1

				o.write('install %s ${RPM_BUILD_ROOT}%s\n' % 
						(makearg(path.basename(file.name)),
						makearg(path.dirname(file.name))))
		else:
			o.write('make install\n')
	o.write('\n')

	# Clean up after BuildRoot
	if not buildroot:
		o.write('#')
	o.write('%clean\n')
	if not buildroot:
		o.write('#')
	o.write('[ "$RPM_BUILD_ROOT" != "/" ] && rm -rf "$RPM_BUILD_ROOT"\n')
	o.write('\n')

	# List all directories and files which are part of package
	o.write('%files\n')
	if options.rpmver >= 2.5:
		o.write('%%defattr(-,%s,%s)\n' % (defaultowner,defaultgroup))

	# Files and directories
	for file in filelist.getfilelist():
		if (file.mode != None and file.mode != defaultmode \
		    and file.mode != defaultexemode) \
		   or (file.owner != None and file.owner != defaultowner \
		       and file.owner != defaultownernum) \
		   or (file.group != None and file.group != defaultgroup \
		       and file.group != defaultgroupnum):
			mode = `file.mode`
			if mode == 'None':
				mode = '-'
			owner = file.owner
			if not owner:
				owner = defaultowner
			group = file.group
			if not group:
				group = defaultgroup
			o.write('%%attr(%s,%s,%s) ' % (mode, owner, group))

		elif options.rpmver < 2.5 and buildroot:
			o.write('%%attr(%s,%s,%s) ' % ('-', defaultowner, defaultgroup))

		if file.isdir:
			o.write('%dir ')	# it's a directory
		else:
			if lookslikeconfigfile(file.name):
				o.write('%config ')	# it's probably a config file
			elif betterlookslikedocfile(file.name):
				o.write('%doc ')	# it's probably a doc file
		o.write(quotefilename(file.name))
		o.write('\n')

	# Extra documentation files
	for file in docfiles:
		if options.rpmver < 2.5 and buildroot:
			o.write('%%attr(%s,%s,%s) ' % ('-', defaultowner, defaultgroup))
		o.write('%%doc %s\n' % quotefilename(file))
	o.write('\n')

	# Add default change log section
	o.write('%changelog\n')
	try:
		userinfo = getpwuid(getuid())
	except KeyError:
		userinfo = ['unknown','',0,0,'','','']
	o.write('* %s %s <%s@%s>\n' % \
		(time.strftime('%a %b %d %Y',time.localtime(time.time())),
		userinfo[4],userinfo[0],uname()[1]))
	o.write('- Initial spec file created by %s with rpm %g compatibility\n'
			% (version_msg, options.rpmver))

	# Flush buffers to guarantee that the following printerr appears last
	o.flush()
	if parseerrs > 0:
		printerr('Warning: %d error(s) parsing install script' % parseerrs)

def usage():
	'Display program usage message and exit with error code'
	sys.stdout.write(version_msg+'\n')
	sys.stdout.write(usage_msg % path.basename(sys.argv[0]))
	sys.exit(1)

def show_license():
	'Display software license'
	sys.stdout.write(version_msg+'\n')
	sys.stdout.write(license)

def autospec():
	'Main command line interface to autospec'

	try:
		optlist, args = getopt.getopt(sys.argv[1:], '?bc:dg:himn:r:zL')
	except getopt.error:
		printerr('Unknown option')
		usage()		# display usage and exit

	if args:
		usage()		# display usage and exit

	options = Options()

	filesmode = 1
	installmode = 0

	# Parse all options
	for opt,arg in optlist:
		if opt == '-h' or \
		   opt == '-?':
			usage()	# display usage and exit

		elif opt == '-L':
			show_license()
			sys.exit(0)

		elif opt == '-b':
			options.binarymode = 1

		elif opt == '-z':
			options.zipmode = 1

		elif opt == '-c':
			options.copyright = arg

		elif opt == '-g':
			options.group = arg

		elif opt == '-d':
			options.docfilepath = None

		elif opt == '-i':
			installmode = 1
			filesmode = 0

		elif opt == '-m':
			options.uselsm = 0

		elif opt == '-n':
			options.rpmnamever = arg

		elif opt == '-r':
			# Ignore everything after second decimal place, e.g. 2.4.12
			simplever = arg
			verlen = find(simplever,'.',find(simplever,'.')+1)
			if verlen >= 0:
				simplever = arg[:verlen]
			try:
				options.rpmver = atof(simplever)
			except ValueError:
				printerr('RPM version not numeric: ' + arg)
				sys.exit(1)

			if options.rpmver > 2.5:
				options.rpmver = 2.5
			elif options.rpmver < 2.4:
				options.rpmver = 2.4

		else:	# should never get here
			printerr('Unknown option ' + opt)
			usage()	# display usage and exit

	if installmode:
		# Write a spec file from output of 'make -n install'
		install2spec(sys.stdin,None,sys.stdout,options)

	elif filesmode:
		# Write a spec file with list of files on stdin
		install2spec(None,sys.stdin,sys.stdout,options)

	sys.exit(0)

if __name__=='__main__':
	autospec()
